import base64
import json
import random
import string
import time
import warnings
from threading import Thread
from typing import Dict, Optional
from urllib.parse import urljoin

from websocket import (  # type: ignore
    WebSocket,
    WebSocketTimeoutException,
    create_connection,
)

from .error import UhliveError, UnknownEvent
from .events import Event

HEARTBEAT_INTERVAL = 15


class Heartbeat(Thread):
    def __init__(self, interval: int, ws: WebSocket) -> None:
        self.ws = ws
        self.interval = interval
        self._running = True
        super().__init__()
        self.daemon = True  # Don't survive main thread

    def run(self) -> None:
        while self._running:
            self.ws.ping("")
            time.sleep(self.interval)

    def stop(self) -> None:
        self._running = False


class Client(object):
    def __init__(self, url: str, identifier: str, token: str, timeout: int = 3):
        """A Client allows to communicate with the Uhlive API.

        :param str url: URL to the uhlive API.
        :param str token: A string specifying an access token.
        :param int timeout: The amount of seconds the session should wait before timing out, defaults to 3
        """
        self.url = urljoin(url, "socket/websocket") + f"?token={token}"
        self.identifier = identifier
        self.timeout = timeout
        self.ref = 0
        self.conversation: Optional[str] = None
        self.heartbeat = None

    def connect(self):
        """Connect to the Uhlive API"""
        self.ws = create_connection(
            self.url, timeout=self.timeout, enable_multithread=True
        )
        self.heartbeat = Heartbeat(HEARTBEAT_INTERVAL, self.ws)
        self.heartbeat.start()

    def join_conversation(
        self,
        conversation_id: str,
        model="fr",
        country="fr",
        readonly: bool = False,
        speaker: str = None,
        interim_results=True,
        rescoring=True,
        origin=0,
    ) -> str:
        """Join a conversation.

        * ``conversation_id``: the identifier of the conversation you want to join.
        * ``readonly``: if you are not going to stream audio, set it to ``True``.
        * ``speaker``: your alias in the conversation, to identify you and your events.
        * ``model``: (if ``readonly`` is ``False``) the ASR language model to be use to recognize
                     the audio you will stream.
        * ``country``: the iso 2 letter country code of the place where the speaker is.
        * ``interim_results``: (``readonly`` = ``False`` only) should the ASR trigger interim result events?
        * ``rescoring``: (``readonly`` = ``False`` only) should the ASR refine the final segment
                         with a bigger Language Model?
                         May give slightly degraded results for very short segments.
        """
        assert (
            readonly or model != ""
        ), "if readonly is false, you must specify a model!"
        self.conversation = f"conversation:{self.identifier}@{conversation_id}"
        if not speaker:
            speaker = self._random_speaker()
        self._send_message(
            "phx_join",
            {
                "readonly": readonly,
                "speaker": speaker,
                "model": model,
                "country": country,
                "interim_results": interim_results,
                "rescoring": rescoring,
                "origin": origin,
            },
        )
        reply = self._read_message()
        if reply["payload"]["status"] != "ok":
            reason = reply["payload"]["response"]["reason"]
            raise UhliveError(f"Cannot join conversation: {reason}")
        return speaker

    def leave_conversation(self) -> None:
        """Leave the current conversation.

        It's a good idea to leave a conversation and continue to consume messages
        until you receive a SpeakerLeft event for your speaker, before you
        close the connection. Otherwise, you may miss parts of the transcription.
        """
        self._send_message("phx_leave", {})

    def send_audio_chunk(self, audio_chunk: bytes):
        """Send an audio chunk (when streaming.)."""
        encoded_audio_chunk = base64.standard_b64encode(audio_chunk).decode("ascii")
        payload = {"blob": encoded_audio_chunk}
        self._send_message("audio_chunk", payload)

    def get_event(self) -> Optional[Event]:
        """Block until an event is received (or the timeout expires).

        Return an event object or None if timeout expired.
        May raise an UhliveError if a network error occurs.
        """
        while True:
            try:
                message = self._read_message()
                # Ignore Phoenix reply events
                # `phx_close` is sent too early by Phoenix, before the channel is actually closed.
                # So users should rely on the application `speaker_left` event matching themselves instead.
                if (
                    message["event"] == "phx_reply"
                    and message["payload"]["status"] == "ok"
                    or message["event"] == "phx_close"
                ):
                    continue  # Ignore Phoenix reply events
                return Event.from_message(message)
            except UnknownEvent as unk:
                warnings.warn(
                    f"Unknown event {unk.event}! You should upgrade your uhlive SDK.",
                    stacklevel=2,
                )
            except WebSocketTimeoutException:
                return None

    def close(self) -> None:
        """Close the connection to the API.

        You'd better `leave_conversation` before."""
        if self.heartbeat:
            self.heartbeat.stop()
        self.ws.close()

    def _random_speaker(self) -> str:
        letters = string.ascii_lowercase
        return "".join(random.choices(letters, k=8))

    def _read_message(self) -> Dict:
        message = self.ws.recv()
        if not message:
            raise UhliveError("Connection is closed, please reconnect")

        return json.loads(message)

    def _send_message(self, event: str, payload: Dict = {}) -> None:
        message = {
            "topic": self.conversation,
            "ref": self.ref,
            "event": event,
            "payload": payload,
        }
        json_message = json.dumps(message).encode("utf8")
        self.ws.send(json_message)
        self.ref += 1
