"""Event definitions."""

from .error import UhliveError, UnknownEvent
from .human_datetime import human_datetime


class Word(dict):
    """Timestamped word."""

    @property
    def start(self):
        """Start time as Unix timestamp in millisecond, according to audio timeline."""
        return self["start"]

    @property
    def end(self):
        """End time as Unix timestamp in millisecond, according to audio timeline."""
        return self["end"]

    @property
    def length(self):
        """Word length in millisecond, according to audio timeline."""
        return self["length"]

    @property
    def word(self):
        """Transcript token string for this word."""
        return self["word"]

    @property
    def confidence(self):
        return self["confidence"]


class Event(object):
    """The base class of all events."""

    def __init__(self, payload) -> None:
        self.payload = payload

    def __repr__(self):
        return f"{self.__class__.__name__}(payload={self.payload})"

    @property
    def speaker(self):
        """The speaker whose speech triggered this event.

        All events are relative to a speaker."""
        return self.payload["speaker"]

    @staticmethod
    def from_message(message):
        """Private method to instantiate the right type of event from the raw websocket message."""
        event = message["event"]
        if event in EVENT_MAP:
            return EVENT_MAP[event](payload=message["payload"])
        if event == "phx_reply" and message["payload"]["status"] == "error":
            raise UhliveError(message["payload"]["response"]["reason"])
        if event == "phx_error":
            raise UhliveError("Server error, channel crashed!")
        raise UnknownEvent(event)


class TimeScopedEvent(Event):
    """Base class for events that are anchored to the audio time line."""

    @property
    def start(self):
        """Start time as Unix timestamp in millisecond, according to audio timeline."""
        return self.payload["start"]

    @property
    def end(self):
        """End time as Unix timestamp in millisecond, according to audio timeline."""
        return self.payload["end"]

    @property
    def length(self):
        """Event length in millisecond, according to audio timeline."""
        return self.payload["length"]


class SpeechDecoded(TimeScopedEvent):
    """The base class of all transcription events."""

    @property
    def transcript(self):
        """Get the transcript of the whole segment as a string"""
        return self.payload["transcript"]

    @property
    def lang(self):
        """Natural Language of the speech.

        As ISO 639-1 code.
        """
        return self.payload["lang"]

    @property
    def country(self):
        """Country location of speaker.

        As ISO 3166-1 code.
        """
        return self.payload["country"]

    @property
    def utterance_id(self):
        """The Utterance id identifies the speech utterance this event transcribes."""
        return self.payload["utterance_id"]

    @property
    def words(self):
        """Get the transcript of the whole segment as a list of timestamped words."""
        return [Word(w) for w in self.payload["words"]]

    def __str__(self):
        return f"[{self.speaker} — {human_datetime(self.start)}] {self.transcript}"


class WordsDecoded(SpeechDecoded):
    """Interim segment transcript event."""

    pass


class SegmentDecoded(SpeechDecoded):
    """Final segment transcript event."""

    pass


class SpeakerJoined(Event):
    @property
    def timestamp(self):
        return self.payload["timestamp"]

    @property
    def interim_results(self):
        return self.payload["interim_results"]

    @property
    def rescoring(self):
        return self.payload["rescoring"]

    @property
    def lang(self):
        """Natural Language of the speech.

        As ISO 639-1 code.
        """
        return self.payload["lang"]

    @property
    def country(self):
        """Country location of speaker.

        As ISO 3166-1 code.
        """
        return self.payload["country"]


class SpeakerLeft(Event):
    """Event emitted by the associated speaker when they left the conversation."""

    @property
    def timestamp(self):
        return self.payload["timestamp"]


class EntityFound(TimeScopedEvent):
    """The base class for all entity annotation events."""

    @property
    def lang(self):
        """Natural Language of the interpretation.

        As ISO 639-1 code.
        """
        return self.payload["lang"]

    @property
    def country(self):
        """Country location of speaker.

        As ISO 3166-1 code.
        """
        return self.payload["country"]

    @property
    def canonical(self):
        """The well formatted form of the entity in the language (string)."""
        return self.payload["annotation"].get("canonical")

    @property
    def original(self):
        """The transcript excerpt that was interpreted, as string."""
        return self.payload["annotation"]["original"]

    @property
    def value(self):
        """The interpreted value in machine understandable form.

        The exact type depends on the entity.
        """
        return self.payload["annotation"].get("value")

    @property
    def confidence(self):
        return self.payload["confidence"]

    def __repr__(self):
        return " ".join(
            (
                f"{self.__class__.__name__} in {self.speaker}: {self.canonical or self.original}",
                f"({self.value})" if self.value != self.canonical else "",
                f"[confidence: {self.confidence:.2f}]",
            )
        )


class EntityReference:
    """Reference to unique Entity in conversation."""

    kind: type
    speaker: str
    start: int

    def __init__(self, entity_type: type, speaker: str, start: int) -> None:
        self.kind = entity_type
        self.speaker = speaker
        self.start = start

    def __repr__(self) -> str:
        return f"Ref({self.kind.__name__} by {self.speaker} at {human_datetime(self.start)})"


class RelationFound(TimeScopedEvent):
    """The base class for all Relation events."""

    @property
    def lang(self):
        """Natural Language of the interpretation.

        As ISO 639-1 code.
        """
        return self.payload["lang"]

    @property
    def confidence(self):
        return self.payload["confidence"]

    @property
    def members(self):
        m = []
        speaker = self.speaker
        for ref in self.payload["members"]:
            kind = EVENT_MAP.get(ref["entity"])
            if kind is not None:
                m.append(EntityReference(kind, speaker, ref["start"]))
        return m

    def __repr__(self):
        return f"{self.__class__.__name__} for {self.members} [confidence: {self.confidence:.2f}]"


class EntityNumberFound(EntityFound):
    """Occurs when a number (formal or quantity) is detected in the transcript.

    The canonical form is the digit form.
    If it's a decimal number, the decimal separator is localized (e.g. "12,5").

    The value is returned as a float.
    """

    pass


class EntityOrdinalFound(EntityFound):
    """Occurs when an ordinal is detected in the transcript.

    The canonical form is the digit form (e.g "22nd").

    The value is its rank returned as a float (e.g. "22.0").
    """

    pass


class RelationCreditCardFound(RelationFound):
    """Link Credit Card entities related to the same real world card."""

    pass


EVENT_MAP = {
    "words_decoded": WordsDecoded,
    "segment_decoded": SegmentDecoded,
    "speaker_joined": SpeakerJoined,
    "speaker_left": SpeakerLeft,
    "entity_number_found": EntityNumberFound,
    "entity_ordinal_found": EntityOrdinalFound,
    "relation_credit_card_found": RelationCreditCardFound,
}


NER_EVENTS = {
    "EntityLocationDeliveryPointFound": "entity_location_delivery_point_found",
    "EntityCreditCardNumberFound": "entity_credit_card_number_found",
    "EntityCreditCardExpirationFound": "entity_credit_card_expiration_found",
    "EntityCreditCardVerificationValueFound": "entity_credit_card_verification_value_found",
    "EntityIbanFound": "entity_iban_found",
    "EntityTelephoneNumberFound": "entity_telephone_number_found",
    "EntityPersFirstnameFound": "entity_pers_firstname_found",
    "EntityPersLastnameFound": "entity_pers_lastname_found",
    "EntityPersSpelledFirstnameFound": "entity_pers_spelled_firstname_found",
    "EntityPersSpelledLastnameFound": "entity_pers_spelled_lastname_found",
    "EntityEmailFound": "entity_email_found",
    "EntityLocationStreetFound": "entity_location_street_found",
    "EntityLocationPostalCodeFound": "entity_location_postal_code_found",
    "EntityLocationCityFound": "entity_location_city_found",
    "EntityLocationCountryFound": "entity_location_country_found",
    "EntityLocationMiscFound": "entity_location_misc_found",
    "EntityHealthOrganFound": "entity_health_organ_found",
    "EntityHealthDiseaseFound": "entity_heath_disease_found",
}


for klass, event_name in NER_EVENTS.items():
    event_type = type(klass, (EntityFound,), {})
    # mypy is confused by dynamic types
    EVENT_MAP[event_name] = event_type  # type: ignore
    globals()[klass] = event_type  # So users can import the class
